///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////  baseskills.c - Implementierung der Klasse der grundlegenden Spielerkommandos

#include "baseskills.h"
#include <stream.h>

BaseSkills::BaseSkills (Comm* communication)
{
	if (communication!=NULL)
		m_pcComm = communication;
	m_bEntered = 0;
	m_sSide = ' ';
	m_nPlayerNumber = 0;
	strcpy (m_sPlaymode, "before_kick_off");
	strcpy (m_sTeamName, "Humbolz");
}

BaseSkills::~BaseSkills ()
{
}


int BaseSkills::EnterField (char* teamname)
{
char message [MAXMESG];

	if (!m_bEntered)
	{
		sprintf (message, "(init %s)", teamname);
		if (m_pcComm->SendMessage (message)==SEND_SUCCESS)
		{
			if (m_pcComm->WaitForMessage (message, MAXMESG, 10, 0)==RECV_SUCCESS)
			{
				if (strcmp (message, "(error no_more_team_or_player)")==0)
					return ENTER_ERROR;
				sscanf (message, "(init %c %i %[^)]", &m_sSide, &m_nPlayerNumber, m_sPlaymode);
				strcpy (m_sTeamName, teamname);
				m_bEntered = 1;
				return ENTER_SUCCESS;
			}
			else
				return ENTER_ERROR;
		}
		else
			return ENTER_ERROR;
	}
	else
		return ENTER_ERROR;
}

int BaseSkills::ReEnterField (char* teamname, int playernumber)
{
char message [MAXMESG];

	if (!m_bEntered)
	{
		sprintf (message, "(reconnect %s %i)",  teamname, playernumber);
		if (m_pcComm->SendMessage (message)==SEND_SUCCESS)
		{
			if (m_pcComm->WaitForMessage (message, MAXMESG, 10, 0)==RECV_SUCCESS)
			{
            if (strcmp (message, "(error can_not_reconnect)")==0)
            {
#ifdef DEBUG 
               cout << "Fehler beim Zurueckmelden!" << endl;
#endif           
					return REENTER_ERROR;
            }              
            sscanf (message, "(reconnect %c %[^)]", &m_sSide, m_sPlaymode);
				m_nPlayerNumber = playernumber;
				strcpy (m_sTeamName, teamname);
				m_bEntered = 1;
				return REENTER_SUCCESS;
			}
			else
				return REENTER_ERROR;
		}
		else
			return REENTER_ERROR;
	}
	else
		return REENTER_ERROR;
}

int BaseSkills::LeaveField ()
{
	if (m_bEntered)
	{
		if (m_pcComm->SendMessage ("(bye)")==SEND_SUCCESS)
		{
			m_bEntered = 0;
			return LEAVE_SUCCESS;
		}
		else
			return LEAVE_ERROR;
	}
	else
		return LEAVE_ERROR;
}

int BaseSkills::MoveTo (float x, float y)
{
char message [MAXMESG];	
	
	if (m_bEntered)
	{
		sprintf (message, "(move %f %f)", x, y);
		if (m_pcComm->SendMessage (message)==SEND_SUCCESS)
			return COMMAND_SUCCESS;
		else
			return COMMAND_ERROR;
	}
	else
		return NOT_ENTERED;
}

int BaseSkills::Turn (float angle)
{
  char message [MAXMESG];	
  
  if (m_bEntered)
    {
      if (angle<-180.0)
	angle += 360.0;
      if (angle>180.0)
	angle -= 360.0;
      if (angle>=-180.0 && angle<=180.0)
	{
	  sprintf (message, "(turn %f)", angle);
	  if (m_pcComm->SendMessage (message)==SEND_SUCCESS)
	    return COMMAND_SUCCESS;
	  else
	    return COMMAND_ERROR;
	}
      else
	return VALUE_OUT_OF_RANGE;
    }
  else
    return NOT_ENTERED;
}

int BaseSkills::Dash (float power)
{
  char message [MAXMESG];	
  
  if (m_bEntered)
    {
      if (power<-30.0)
	power = -30.0;
      if (power>100.0)
	power = 100.0;      
      sprintf (message, "(dash %f)", power);
      if (m_pcComm->SendMessage (message)==SEND_SUCCESS)
	return COMMAND_SUCCESS;
      else
	return COMMAND_ERROR;
	
    }
  else
    return NOT_ENTERED;
}

int BaseSkills::Kick (float power, float angle)
{
char message [MAXMESG];	
	
	if (m_bEntered)
	{
		if (angle<-180.0)
			angle += 360.0;
		if (angle>180.0)
			angle -= 360.0;
		if (power<-30.0)
			power = -30.0;
		if (power>100.0)
			power = 100.0;

		if (angle>=-180.0 && angle<=180.0)
		{
			sprintf (message, "(kick %f %f)", power, angle);
			if (m_pcComm->SendMessage (message)==SEND_SUCCESS)
				return COMMAND_SUCCESS;
			else
				return COMMAND_ERROR;
		}
		else
			return VALUE_OUT_OF_RANGE;
	}
	else
		return NOT_ENTERED;
}

int BaseSkills::Say (char* message)
{
char mess [MAXMESG];

	if (m_bEntered)
	{
		sprintf (mess, "(say %s)", message);
		if (m_pcComm->SendMessage (mess)==SEND_SUCCESS)
			return COMMAND_SUCCESS;
		else
			return COMMAND_ERROR;
	}
	else
		return NOT_ENTERED;
}

int BaseSkills::ChangeView (int width, int quality)
{
char message [MAXMESG];
char widthstr [20];
char qualstr [20];

	if (m_bEntered)
	{
		switch (width)
		{
			case VIEW_WIDTH_NARROW:
				strcpy (widthstr, "narrow");
				break;
			case VIEW_WIDTH_NORMAL:
				strcpy (widthstr, "normal");
				break;
			case VIEW_WIDTH_WIDE:
				strcpy (widthstr, "wide");
				break;
			default:
				return COMMAND_ERROR;
		}
		switch (quality)
		{
			case VIEW_QUAL_HIGH:
				strcpy (qualstr, "high");
				break;
			case VIEW_QUAL_LOW:
				strcpy (qualstr, "low");
				break;
			default:
				return COMMAND_ERROR;
		}
		sprintf (message, "(change_view %s %s)", widthstr, qualstr);
		if (m_pcComm->SendMessage (message)==SEND_SUCCESS)
			return COMMAND_SUCCESS;
		else
			return COMMAND_ERROR;
	}
	else
		return NOT_ENTERED;
}

char BaseSkills::GetSide ()
{
	return m_sSide;
}

char* BaseSkills::GetPlaymode ()
{
	return m_sPlaymode;
}

int BaseSkills::GetPlayerNumber ()
{
	return m_nPlayerNumber;
}
