#include "entscheidung.h"
#include "referee.h"
#include <math.h>
#include <stdlib.h>


// Jan

/**
 * Gibt 1 zurueck wenn der Abstand der zwei Punkte <= NEAR_RANGE ist, sonst 0
 */
  int Brain::isNear(Punkt point, Punkt otherPoint) {
    double distance;
    distance =  (point - otherPoint).norm();
 
    switch(m_cWeltmodell->frage_rolle()) {
    case ROLLE_TORWART:
      if (distance < NEAR_RANGE) return 1;
      else return 0;
      break;
    case ROLLE_LINKER_VERTEIDIGER:
    case  ROLLE_RECHTER_VERTEIDIGER:
      if (distance < 3.0) return 1;
      else return 0;
      break;
    default:
      if (distance < 3.0) return 1;
      else return 0;
      break;
    }
  }

/**
 * Berechnet den neuen Status: was aus der Situation heraus am besten gemacht
 * werden sollte.
 */
  int Brain::decideNew() {
    
    Ball ball = m_cWeltmodell->frage_ball();
    Spieler_Selbst self = m_cWeltmodell->frage_selbst();
    Punkt ballPos = ball.frage_p();
    Punkt selfPos = self.frage_p();
    Koerper distanceToBall = (ballPos-selfPos).norm();
    Zeit time = self.frage_jetzt();

#ifdef DEBUG
    cout << "decideNew: distanceToBall: " << distanceToBall << endl;
#endif

    if(ball.frage_wert() >= 3) {
#ifdef DEBUG
      cout << "decideNew: Ich suche erst mal den Ball" << endl;
#endif
      // nur mit sicheren Informationen neue Entscheidung treffen;
      lookNewTime = time+1;
      Winkel turnAngle = m_cAdvSkills->ComputeTurnAngle(ballPos);
      if(turnAngle > -40.0 && turnAngle < 40.0)
	return STATE_SEEK_BALL;
      else
	return STATE_LOOK_FOR_BALL;
    }
    else {
      if(distanceToBall >= THINK_RANGE) {
#ifdef DEBUG
	cout << "decideNew: Ich will nach Hause laufen" << endl;
#endif
	// Ball ist sehr weit weg -> ich begebe mich nach Hause wenn;
	// ich nicht schon dort bin, ansonsten drehe ich mich genau zum Ball;

	interceptPos = m_cWeltmodell->frage_stammpos_spiel();
	if( isNear(selfPos, interceptPos) ) { 
	  lookNewTime = time+1;
	  return STATE_TURN_TO_BALL;
	}
	else {
	  minSteps = (long)(distanceToBall - THINK_RANGE)/2;
	  if(minSteps > 10) minSteps = 10; // maximal 10 Schritte auf die Heimatposition zu halten;
	  lookNewTime = time + minSteps;
	  return STATE_TAKE_REGULAR_POS;     
	}
      }
      else {
	// distanceToBall < THINK_RANGE;

	computeForAllStepsToBall();
	// Jetzt ist ueber jeden Spieler bekannt wieviel Schritte er zum;
	// Ball braucht. Desweiteren sind die Spieler jedes Teams bekannt;
	// die am schnellsten am Ball sein koennen. Die Anzahl der Schritte;
	// nach denen ein Spieler den Ball erreichen kann ist auch bekannt;
	
	// egal fuer was ich mich entscheide, ich suche erst wieder nach;
	// den Ball, wenn ihn auch jemand erreicht haben kann;
	// -> minSteps wird zur aktuellen Zeit addiert um lookNewTime zu erhalten;

	if(minStepsOur == PLAYER_SELF) {
#ifdef DEBUG
	  cout << "decideNew: Ich will zum Ball laufen" << endl;
#endif
	  // ich bin naechster am Ball -> zur Position laufen, wo der Ball liegen wird;

	  // der Ball ist noch ein wenig entfernt -> ungefaehr Abfangpunkt ausrechnen;
	  Ball dummyBall = ball; // eine Kopie des Balles ziehen;
	  lookNewTime = time + minSteps;

	  if(stepsOurPlayer[minStepsOur] <= 3) {
	    // ich bin sehr nah am Ball -> genau ausrechnen wo der Ball abzufangen ist;
	    dummyBall.simulieren(minSteps); // Position berechnen auf der Ball sein wird;
	    interceptPos = dummyBall.frage_p();
#ifdef DEBUG
	    cout << "SimulatedSteps: " << minSteps << endl;
	    cout << "InterPos:  x: " << dummyBall.frage_p().x1 << " y: " << dummyBall.frage_p().x2 << endl;
#endif
	    return STATE_FETCH_NEAR_BALL;
	  }
	  else {
	    dummyBall.simulieren(minSteps + FREE_STEPS); // Position berechnen auf der Ball sein wird;
	    interceptPos = dummyBall.frage_p();
#ifdef DEBUG
	    cout << "SimulatedSteps: " << minSteps << endl;
	    cout << "InterPos:  x: " << dummyBall.frage_p().x1 << " y: " << dummyBall.frage_p().x2 << endl;
#endif
	    return STATE_FETCH_BALL;
	  }
	}
	else {
	  // ich bin nicht der naechste, bin aber vielleicht im Einflussbereich des Balles;
	  int influenceSelf;

	  if(minStepsOur == PLAYER_UNKNOWN)
	    influenceSelf = 0;
	  else
	    {
	      // der Spieler von uns der am naechsten am Ball ist;
	      Spieler player = m_cWeltmodell->frage_spieler(minStepsOur, team_wir);
	      
	      // berechnen ob ich mich in der Einflussposition befinde;
	      if(stepsOurPlayer[minStepsOur] <= INFLUENCE_PLAYER_STEPS)
		influenceSelf = neighbours(player, (Spieler)self);
	      else
		influenceSelf = 0;
	    }
	  
	  if(influenceSelf) {
#ifdef DEBUG
	    cout << "decideNew: Ich bin in der Einflussposition des Balles" << endl;
#endif
	    interceptPos = computeBestPosToWaitForBall(minStepsOur);	    
	    if( isNear(selfPos, interceptPos) ) { 
	      lookNewTime = time+1;
	      return STATE_TURN_TO_BALL;
	    }
	    else {
//	      if(minSteps < 10) minSteps = 10; // mindestens 10 Schritte auf die Einflussposition zu halten;
	      lookNewTime = time + minSteps;
	      return STATE_TAKE_INFLUENCE_POS;
	    }
	  }
	  else {
#ifdef DEBUG
	    cout << "decideNew: Ich will nach Hause laufen" << endl;
#endif
	    interceptPos = m_cWeltmodell->frage_stammpos_spiel();	    
	    if( isNear(selfPos, interceptPos) ) { 
	      lookNewTime = time+1;
	      return STATE_TURN_TO_BALL;
	    }
	    else {
//	      if(minSteps < 10) minSteps = 10; // mindestens 10 Schritte auf die Heimatposition zu halten;
	      lookNewTime = time + minSteps;
	      return STATE_TAKE_REGULAR_POS;     
	    }
	  }
	  
	}


//	lookNewTime = time + (long)(distanceToBall/2);
//	interceptPos = ballPos;
//	return STATE_FETCH_BALL;	
	
      }  
    }
//    assert(0); // sollte nie eintreten (irgend ein Fall enthaelt kein return);
  }


/**
 * Berechnet fuer alle Spieler wie lange sie zum Ball brauchen. Dabei werden
 * die globalen Arrays stepsOurPlayer und stepsTheirPlayer gefuellt.
 * Desweiteren werden die Spieler jedes Teams bestimmt, die am schnellsten
 * beim Ball sein koennen (minStepsOur und minStepsTheir). Die minimale
 * Schrittanzahl in denen der Ball von irgend jemanden erreicht werden
 * kann wird in die globale Variable minSteps gespeichert.
 */
  void Brain::computeForAllStepsToBall() {
    int i;

    // Initierung der Minimalwerte;
    minStepsOur = PLAYER_UNKNOWN;
    minStepsTheir = PLAYER_UNKNOWN;
    
    // Schritte fuer mich selbst berechnen;
    stepsOurPlayer[PLAYER_SELF] = computeStepsToBall(team_wir, PLAYER_SELF);

    if(stepsOurPlayer[PLAYER_SELF] < stepsOurPlayer[minStepsOur]){
      
      minStepsOur = PLAYER_SELF;
    }
    Spieler_Selbst playerMy = m_cWeltmodell->frage_selbst();
#ifdef DEBUG
    cout << "MYSTAMINA: " << playerMy.frage_stamina() << endl;
#endif
    
    // die Schritte fuer meine Teamkameraden;
    for(i=0;i<10;i++) {
      stepsOurPlayer[i] = computeStepsToBall(team_wir, i);
  
      if(stepsOurPlayer[i] < stepsOurPlayer[minStepsOur])
	minStepsOur = i;
    }
    
    
    // die Schritte fuer die gegenerischen Spieler;
    for(i=0;i<11;i++) {
      stepsTheirPlayer[i] = computeStepsToBall(team_sie, i);
      
      if(stepsTheirPlayer[i] < stepsTheirPlayer[minStepsTheir])
	minStepsTheir = i;
    }
    
    #ifdef DEBUG
    cout << "computeForAllStepsToBall: OurMinPlayer: " << minStepsOur << "   TheirMinPlayer: " << minStepsTheir << endl;
    #endif
    
    // die minimalen Schritte nachdem jemand beim Ball sein kann;
    if(stepsOurPlayer[minStepsOur] <= stepsTheirPlayer[minStepsTheir])
      minSteps = stepsOurPlayer[minStepsOur];
    else
      minSteps = stepsTheirPlayer[minStepsTheir];
    
    #ifdef DEBUG
    cout << "computeForAllStepsToBall: MinSteps: " << minSteps << endl;
    #endif
    
    return;
    
  }

/**
 * Berechnet die Schritte die der angegebene Spieler braucht um 
 * zum Ball zu gelangen
 */
  int Brain::computeStepsToBall(int team, int player, int freeSteps = 0) {
    Spieler_Selbst dummyPlayer;
    double dummyStamina;
    int steps;


    if(team == team_wir && player == PLAYER_SELF)
      dummyPlayer = m_cWeltmodell->frage_selbst();
    else {  
      dummyPlayer = ((Spieler_Selbst&) m_cWeltmodell->frage_spieler(player, team));
      if(&dummyPlayer == NULL || dummyPlayer.ist_unsicher()) return MAX_STEPS+1;
//      dummyPlayer.setze_richtung(dummyPlayer.frage_v());
      Vektor playerDirection;
      playerDirection.x1 = 1.0;
      playerDirection.x2 = 0.0;
      dummyPlayer.setze_richtung (playerDirection);
    }

    if( (m_cWeltmodell->frage_ball().frage_p() - dummyPlayer.frage_p()).norm() >= THINK_RANGE )
      return MAX_STEPS+1;

    dummyStamina = dummyPlayer.frage_stamina();

    double tempTurn, tempLastDash, tempStamina;
    Ball ball = m_cWeltmodell->frage_ball();
    ball.simulieren(freeSteps);
    
    steps = m_cAdvSkills->BestWay2Ball(ball, dummyPlayer, 
				       &tempTurn, &tempLastDash, &tempStamina);

    if(steps == -1) steps = MAX_STEPS+1;
    #ifdef DEBUG
    cout << "computeStepsToBall: Steps ";
    if(team == team_wir)
      if(player == PLAYER_SELF) cout << "Our Self: ";
      else cout << "Our Nr." << player << ": ";
    else cout << "Their Nr." << player << ": ";
    cout << steps;
    #endif

// Jan:  Hier wieder die Stamina einbauen wenn das Update funktioniert;    
//    if(dummyStamina < tempStamina)
//      steps += (int) (tempStamina-dummyStamina) / STAMINA_INC +1;

    if(steps > MAX_STEPS+1) steps = MAX_STEPS+1;

    #ifdef DEBUG
    cout << "   real: " << steps << endl;
    #endif

    return steps;
    
  }

/**
 * Testet ob die angegeben Spieler Nachbarn sind. Die Methode liefert dann
 * 1 zurueck, sonst 0.
 */
  int Brain::neighbours(Spieler playerWithBall, Spieler playerInfluence) {
    double dist;
    
    dist = (playerWithBall.frage_p() - playerInfluence.frage_p()).norm();
    if(dist < INFLUENCE_RANGE) return 1;
    else return 0;
  }

/** 
 * Berechnet fuer einen Spieler der sich in der Einflussposition des Balls
 * befindet, die Position auf der er am besten angespielt werden kann
 * und gleichzeitig mindestens einen anderen seiner Freunde wieder gut
 * anspielen kann
 */
#define INFLUENCE_POS_INC 4.0
#define INFLUENCE_POS_STEPS 5

  Punkt Brain::computeBestPosToWaitForBall(int playerNumber) {
    Punkt point, tempPoint;
    double quality;
    double tempQuality;
    double x, y;
    int i, j;
    
    return m_cWeltmodell->frage_stammpos_spiel();

//    point = m_cWeltmodell->frage_selbst().frage_p();
    point = m_cWeltmodell->frage_stammpos_spiel();
    x = point.x1 - (double)(INFLUENCE_POS_STEPS/2)*INFLUENCE_POS_INC;
    y = point.x2 - (double)(INFLUENCE_POS_STEPS/2)*INFLUENCE_POS_INC;
    
#ifdef DEBUG
    Spieler player = m_cWeltmodell->frage_spieler(playerNumber, team_wir);
    cout << "computeBestPosToWaitForBall: PlayerWithBall: x: " << player.frage_p().x1 << " y: " << player.frage_p().x2 << endl;
#endif

    quality = computeQualityOfPos(point, playerNumber);
    cout << "computeBestPosToWaitForBall: Pos: x: " << point.x1 << " y: " << point.x2 << " Quality: " << quality << endl;

    for(i=0; i<INFLUENCE_POS_STEPS; i++) {
      for(j=0; j<INFLUENCE_POS_STEPS; j++) {
	tempPoint.x1 = x + i*INFLUENCE_POS_INC;
	tempPoint.x2 = y + j*INFLUENCE_POS_INC;
	
	tempQuality = computeQualityOfPos(tempPoint, playerNumber);

#ifdef DEBUG
	cout << "computeBestPosToWaitForBall: Pos: x: " << tempPoint.x1 << " y: " << tempPoint.x2 << " Quality: " << tempQuality << endl;
#endif
	if(tempQuality > quality) {
	  point = tempPoint;
	  quality = tempQuality;

#ifdef DEBUG
	  cout << "computeBestPosToWaitForBall: New Position is better" << endl;
#endif
	}
      }
    }

#ifdef DEBUG
    cout << endl;
#endif

    return point;    
  }

/**
 * Berechnet die Qualitaet der angegeben Position. Qualitaet bedeutet
 * hier wie frei der eigene Spieler zu dem Spieler mit den Ball und
 * mindestens einen anderen Spieler steht. Um so weniger gegnerische
 * Spieler dazwischen laufen koennen um so freier steht der eigene
 * Spieler und um so hoeher ist die Qualitaet der Position.
 */
  double Brain::computeQualityOfPos(Punkt point, int playerNumber) {
  
    double source, dest, maxDest, sumOtherDest, quality;
    int i;
    Spieler player;
 
    source = computeFreedom(point, playerNumber);
    cout << "    computeQualityOfPos: Freiheit Ballquelle: (Nr.:" << playerNumber << ") " << source << endl;

    /*    
    maxDest = 0;
    for(i=0; i<10; i++) {
      if(i != playerNumber) {
	player = m_cWeltmodell->frage_spieler(i, team_wir);
	if(!(&player == NULL || player.ist_unsicher())) {
	  dest = computeFreedom(point, player);
	  cout << "    computeQualityOfPos: Freiheit Ballsenke: (Nr.:" << i << ") " << dest << endl;
	  if(dest > maxDest) {
	    sumOtherDest += maxDest;
	    maxDest = dest;
	  }
	  else {
	    sumOtherDest += dest;
	  }
	}
      }
    }
    */
    
//    quality = source + maxDest + (int) 0.5 * sumOtherDest;
    quality = source;

    
    return quality;
  }

  double Brain::computeFreedom(Punkt point, Spieler player) {
    double distanceToOurPlayer, distanceToTheirPlayer;
    Winkel alpha; Koerper distance, tempDistance;
    Spieler enemyPlayer;
    int i;
    
    distance = 100.0;

    distanceToOurPlayer = (player.frage_p() - point).norm();
      
    for(i=0; i<11; i++) {
      enemyPlayer = m_cWeltmodell->frage_spieler(i, team_sie);
      if(!(&enemyPlayer == NULL || enemyPlayer.ist_unsicher())) {
	distanceToTheirPlayer = (enemyPlayer.frage_p() - point).norm();
	if(distanceToTheirPlayer < distanceToOurPlayer) {

#ifdef DEBUG	  
	  cout << "      computeFreedom: zu bewertenden Punkt: x: " << point.x1 << " y: " << point.x2 << endl;
	  cout << "      computeFreedom: unser Spieler (Ball): x: " << player.frage_p().x1 << " y: " << player.frage_p().x2 << endl;
	  cout << "      computeFreedom: gegnerischer Spieler: x: " << enemyPlayer.frage_p().x1 << " y: " << enemyPlayer.frage_p().x2 << endl;
#endif

	  // Bestimmen des Abstandes des gegenerischen Spielers;
	  // von der Gerade zwischen den Spieler player und dem;
	  // Punkt point;
	  alpha = (player.frage_p() - point).arg() -
	    (enemyPlayer.frage_p() - point).arg();
	  if(alpha > PI) alpha -= 2*PI;
	  if(alpha < -PI) alpha += 2*PI;
	  
	  if(alpha >= -PI/2 && alpha <= PI/2) {
	    // nur wirklich gegenerische Spieler betrachten;
	    // die dazwischen laufen koennen;
	    tempDistance = distanceToTheirPlayer * sin(alpha);
	    if(tempDistance < 0) tempDistance *= -1.0;
#ifdef DEBUG
	    cout << "      computeFreedom: Distanz von Gegner: (Nr.:" << i << ") " << tempDistance << " (Pos: (" << enemyPlayer.frage_p().x1 << ", " << enemyPlayer.frage_p().x2 << "))" << endl;
#endif
	    
	    if(tempDistance < distance) distance = tempDistance;
	  }
	  
	}
      }
    }
    
    return distance;
  }

// Jan
