#ifndef __vektorh__

#define __vektorh__

#include<math.h>
#include<float.h>

#ifndef PI
#define PI		3.141592654
#endif

typedef double Koerper;
typedef double Winkel;

const Koerper Koerper_Max = DBL_MAX;
const Koerper Koerper_Min = DBL_MIN;

const Koerper eps = 0.00001; // vorausgesetzte Maschinengenauigkeit

inline Koerper betrag(Koerper x) {return x < 0 ? -x : x;}
inline Koerper norm(Koerper x,Koerper y) {return sqrt(x * x + y * y);}
inline Winkel deg_in_rad(Winkel gradmass) {return PI * gradmass / 180;}
inline Winkel rad_in_deg(Winkel bogenmass) {return 180 * bogenmass / PI;}
	// Umrechnung zwischen Gradmasen
inline int signum(Koerper zahl) {return zahl < 0 ? -1 : 1;}
inline int ist_ungefaehr(Koerper x,Koerper y, double fehler = eps) {return betrag(x - y) < fehler;}


struct Vektor
	{
	Koerper x1,x2;

/*	inline Vektor(void) {x1 = x2 = 0;}
	inline Vektor(Koerper y1,Koerper y2)	{x1 = y1,x2 = y2;} */

	inline Koerper proj1(void) const {return x1;}
	inline Koerper proj2(void) const {return x2;}

	inline short kollinear(const Vektor &v) const
		// != 0, falls v koll. zu this
		{return proj1() * v.proj2() == v.proj1() * proj2();}

	inline short operator==(const Vektor &v) const
		{return proj1() == v.proj1() && proj2() == v.proj2();}

	inline Vektor &operator=(const Vektor &eq) {x1 = eq.x1,x2 = eq.x2; return *this;}
	
	inline Vektor operator-(void) const
		{Vektor ret; ret.x1 = -x1,ret.x2 = -x2; return ret;}
	inline Vektor operator+=(const Vektor add)
		{x1 += add.x1,x2 += add.x2; return *this;}
	inline Vektor operator-=(const Vektor sub)
		{return (*this += -sub);}
	inline Vektor operator*=(const Koerper skalar)
		{x1 *= skalar,x2 *= skalar; return *this;}
	inline Vektor operator+(Vektor add) const
		{add += *this; return add;}
	inline Vektor operator-(const Vektor sub) const
		{return (*this + (-sub));}
	friend inline Vektor operator*(Koerper skalar,Vektor v); // S-Mul.
	
	inline Koerper operator*(const Vektor v) const // Skalarmultiplikation
		{return (x1 * v.x1 + x2 * v.x2);}
	inline Koerper norm(void) const // Norm des Vektors
		{return sqrt(*this * *this);}

	Vektor normal(void) const; // Berechnung eines Normaleneinheitsvektors	

	void drehen(const Winkel phi);
		// Drehen des Vektors um den Nullpunkt um den Winkel phi
	void drehen_um(const Winkel phi,const Vektor punkt);
		// Drehen des Vektors um den Punkt punkt um den Winkel phi
	void skalieren(const Koerper s1,const Koerper s2);
		// Skalieren eines Vektors mit unterschiedlichen Skalaren fur x1/2-Achse
	void spiegeln_x1(const Koerper s1 = 0);
		// Spiegeln an der x2-Achse bzw. an x1 = s1
	void spiegeln_x2(const Koerper s2 = 0);
		// Spiegeln an der x1-Achse bzw. an x2 = s2
	
	Winkel arg(void) const;
		// Liefert das Argument des Vektors (aus (-pi;pi] )
	void setzen(Winkel phi,Koerper l);
		// Vektor aus Winkel und Lange konstruieren
	};

inline Vektor operator*(Koerper skalar,Vektor v)
	{
	v *= skalar;
	return v;
	}

inline Koerper norm(Vektor v)
	{
	return v.norm();
	}

#endif
