/////////////////////////////////////////////////////////////////////////////////////////
/////  sensors.c - Implementierung der Klassen relpos, player und sensors

#include "sensors.h"
#include <stream.h>
#include <string.h>

/////  Klasse relpos zur Speicherung einer relativen Positionsinformation
RelPos::RelPos ()
{
  m_fDistance = DIST_NOTVALID;
  m_fAngle = ANGLE_NOTVALID;
  m_fDistChange = DISTCHANGE_NOTVALID;
  m_fDirChange = DIRCHANGE_NOTVALID;
  m_bValid = 0;
}

RelPos & RelPos::operator=(RelPos &x)
{
  if (&x!=this)
    {
      m_fDistance = x.m_fDistance;
      m_fAngle = x.m_fAngle;
      m_fDistChange = x.m_fDistChange;
      m_fDirChange = x.m_fDirChange;
      m_bValid = x.m_bValid;
    }
  return (*this);
}

void RelPos::SetDistance (float dist)
{
  m_fDistance = dist;
  m_bValid = 1;
}

void RelPos::SetAngle (float angle)
{
  m_fAngle = angle;
  m_bValid = 1;
}

void RelPos::SetDistChange (float distchange)
{
  m_fDistChange = distchange;
  m_bValid = 1;
}

void RelPos::SetDirChange (float dirchange)
{
  m_fDirChange = dirchange;
  m_bValid = 1;
}

void RelPos::SetAll (float dist, float angle, float distchange, float dirchange)
{
  m_fDistance = dist;
  m_fAngle = angle;
  m_fDistChange = distchange;
  m_fDirChange = dirchange;
  m_bValid = 1;
}

float RelPos::GetDistance ()
{
  if (m_bValid)
    return m_fDistance;
  else
    return DIST_NOTVALID;
}

float RelPos::GetAngle ()
{
  if (m_bValid)
    return m_fAngle;
  else
    return ANGLE_NOTVALID;
}

float RelPos::GetDistChange ()
{
  if (m_bValid)
    return m_fDistChange;
  else
    return DISTCHANGE_NOTVALID;
}

float RelPos::GetDirChange ()
{
  if (m_bValid)
    return m_fDirChange;
  else
    return DIRCHANGE_NOTVALID;
}

void RelPos::GetAll (float* dist, float* angle, float* distchange, float* dirchange)
{
  *dist = m_fDistance;
  *angle = m_fAngle;
  *distchange = m_fDistChange;
  *dirchange = m_fDirChange;
}

int RelPos::IsValid ()
{
  return m_bValid;
}

void RelPos::SetValid (int valid)
{
  m_bValid = valid;
  if (!valid)
    {
      m_fDistance = DIST_NOTVALID;
      m_fAngle = ANGLE_NOTVALID;
      m_fDistChange = DISTCHANGE_NOTVALID;
      m_fDirChange = DIRCHANGE_NOTVALID;
    }
}

/////  Klasse player zur Speicherung der Informationen zu einem Spieler
Player::Player ()
{
  strcpy (m_sTeam, PLAYER_TEAM_NOTVALID);
  m_nNumber = PLAYER_NUMBER_NOTVALID;
  m_bValid = 0;
}

Player::Player (char* team, int number)
{
  SetTeamName (team);
  SetPlayerNumber (number);
  m_bValid = 1;
}

Player::~Player ()
{
}

/*
Player & Player::operator= (Player &x)
{
  if (&x!=this)
    {
      m_fDistance = x.m_fDistance;
      m_fAngle = x.m_fAngle;
      m_nVDirection = x.m_nVDirection;
      m_bValid = x.m_bValid;
      strcpy (m_sTeam, x.m_sTeam);
      m_nNumber = x.m_nNumber;
  }
  return (*this);
}
*/

void Player::SetPlayerNumber (int number)
{
  m_nNumber = number;
  m_bValid = 1;
}

void Player::SetTeamName (char* team)
{
  strcpy (m_sTeam, team);
  m_bValid = 1;
}

int Player::GetPlayerNumber ()
{
  return m_nNumber;
}

char* Player::GetTeamName ()
{
  return m_sTeam;
}

/////  Klasse sensors zur Kapselung aller sensorischer Informationen
Sensors::Sensors ()
{
  m_pcComm = NULL;
  m_nMessageType = MESSAGETYPE_HEAR;
  m_nTime = 0;
}

Sensors::Sensors (Comm* communication)
{
  m_nMessageType = MESSAGETYPE_HEAR;
  m_nTime = 0;
  InitSensorsComm (communication);
}

Sensors::~Sensors ()
{
}

void Sensors::InitSensorsComm (Comm* communication)
{
  m_pcComm = communication;
  m_bOwnComm = 1;
}

int Sensors::ListenAndWatch ()
{
  int res;
  char message [MAXMESG];
  
  if (m_bOwnComm && m_pcComm!=NULL)
    {
      res = m_pcComm->WaitForMessage (message, MAXMESG, 1, 0);
      if (res==RECV_SUCCESS)
	return ParseSensoryData (message);
      else
	return res;
    }
  else
    return 0;
}

int Sensors::ListenAndWatch (int usec)
{
  int res;
  char message [MAXMESG];
  
  if (m_bOwnComm && m_pcComm!=NULL)
    {
      res = m_pcComm->WaitForMessage (message, MAXMESG, 0, usec);
      if (res==RECV_SUCCESS)
	{
	  return ParseSensoryData (message);
	}
      else
	return res;
    }
  else{
    return 0;
  }
}

int Sensors::ParseSensoryData (char* buf)
{
  int i;
  char sensorinfo [MAXMESG];
  char message [MAXMESG];
  
  strcpy (sensorinfo, buf);
  //  Ruecksetzen der Sensoren;
  m_cBall.SetValid (0);
  m_cRightGoal.SetValid (0);
  m_cLeftGoal.SetValid (0);
  m_cAnonymousGoal.SetValid (0);
  for (i=FLAG_LEFTBOTTOM; i<FLAG_NOTVALID; i++)
    m_cFlags [i].SetValid (0);	
  for (i=LINE_LEFT; i<LINE_NOTVALID; i++)
    m_cLines [i].SetValid (0);
  m_cPlayers.MakeEmpty ();
  
  if (strncasecmp (sensorinfo, "(see ", 4)==0) {
    m_nMessageType = MESSAGETYPE_SEE;
    sscanf (sensorinfo, "(see %d%[^\n]", &m_nTime, message);
    ParseSights (message);
  }
  else if (strncasecmp (sensorinfo, "(hear ", 5)==0) {
    m_nMessageType = MESSAGETYPE_HEAR;
    sscanf (sensorinfo, "(hear %d%[^\n]", &m_nTime, message);
    ParseSounds (message);
  }
  else if (strncasecmp (sensorinfo, "(error ", 6)==0) {
    m_nMessageType = MESSAGETYPE_ERROR;
    sscanf (sensorinfo, "(error %[^\n]", m_sError);
  }
  return 1;
}

RelPos Sensors::GetBall ()
{
  return m_cBall;
}

RelPos Sensors::GetRightGoal ()
{
  return m_cRightGoal;
}

RelPos Sensors::GetLeftGoal ()
{
  return m_cLeftGoal;
}

RelPos Sensors::GetAnonymousGoal ()
{
  return m_cAnonymousGoal;
}
	
RelPos Sensors::GetAFlag (int flagid)
{
  if (flagid>=FLAG_LEFTBOTTOM && flagid<FLAG_NOTVALID)
    return m_cFlags [flagid];
  else
    return m_cFlags [FLAG_NOTVALID];
}

RelPos Sensors::GetALine (int lineid)
{
  if (lineid>=LINE_LEFT && lineid<LINE_NOTVALID)
    return m_cLines [lineid];
  else
    return m_cLines [LINE_NOTVALID];
}

ALinkList <Player> *Sensors::GetPlayers ()
{ 
  return &m_cPlayers; 
}

int Sensors::GetMessageType ()
{
  return m_nMessageType;
}

int Sensors::GetTime ()
{
  return m_nTime;
}

float Sensors::GetMessageSender ()
{
  if (m_nMessageType==MESSAGETYPE_HEAR)
    return m_fSender;
  else
    return MESSAGESENDER_NOTVALID;
}

char* Sensors::GetMessage ()
{
  if (m_nMessageType==MESSAGETYPE_HEAR)
    return m_sMessage;
  else
    return NULL;
}

void Sensors::ParseSights (char* message)
{
  float dist, angle;
  float distchange, dirchange;
  char object [MAXMESG];
  
  while (message [0]!=')') {
    dist = DIST_NOTVALID;
    angle = ANGLE_NOTVALID;
    distchange = DISTCHANGE_NOTVALID;
    dirchange = DIRCHANGE_NOTVALID;
    sscanf (message, " (%[^)]) %f%[^\n]", object, &dist, message);
    if (message [0]==')')
      {
	message++; // niedrige Qualitaet, nur Direction wird uebermittelt;
	angle = dist;
	dist = DIST_NOTVALID;
      }
    else
      {
	sscanf (message, " %f%[^\n]", &angle, message);
	if (message [0]==')')
	  message++;
	else
	  {
	    sscanf (message, " %f%[^\n]", &distchange, message);
	    if (message [0]==')')
	      message++;
	    else
	      {
		sscanf (message, "%f%[^\n]", &dirchange, message);
		if (message [0]==')')
		  message++;
	      }
	  }
      }
    strcat (object, ")");
    if (strcasecmp (object, "(ball)")==0)
      SeeBall (dist, angle, distchange, dirchange);
    else if (strncasecmp (object, "(player", 7)==0)
      {
	char team [MAXMESG];
	int number;
	
	sscanf (object, "(player%[^)]", object); // Schluesselwort "(player" abschneiden;
	strcpy (team, PLAYER_TEAM_NOTVALID);
	number = PLAYER_NUMBER_NOTVALID;
	if (object [0]!=')')
	  {
	    sscanf (object, " %s%[^\n]", team, object);
	    if (object [0]!=')')
	      sscanf (object, " %i%[^\n]", &number);
	  }				
	SeePlayer (dist, angle, distchange, dirchange, team, number);
      }
    else
      SeeMarker (object, dist, angle, distchange, dirchange);
  }
}

void Sensors::ParseSounds (char* message)
{
  char sender [MAXMESG];
  
  sscanf (message, " %s %[^)]", sender, m_sMessage);
  if (!strcmp (sender, "referee"))
    m_fSender = MESSAGESENDER_REFEREE;
  else
    if (!strcmp (sender, "self"))
      m_fSender = MESSAGESENDER_SELF;
    else
      m_fSender = atof (message);	
}

void Sensors::SeeMarker (char* marker, float dist, float angle, float distchange, float dirchange)
{
  //  Erkennen der Tore;
  if (strcasecmp (marker, "(goal r)")==0)
    {
      m_cRightGoal.SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(goal l)")==0)
    {
      m_cLeftGoal.SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(goal)")==0)
    {
      m_cAnonymousGoal.SetAll (dist, angle, distchange, dirchange);
      return;
    }
  
  //  Erkennen der aeusseren Flaggen;
  if (strcasecmp (marker, "(flag l b)")==0)
    {
      m_cFlags [FLAG_LEFTBOTTOM].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag l t)")==0)
    {
      m_cFlags [FLAG_LEFTTOP].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag c t)")==0)
    {
      m_cFlags [FLAG_CENTERTOP].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag r t)")==0)
    {
      m_cFlags [FLAG_RIGHTTOP].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag r b)")==0)
    {
      m_cFlags [FLAG_RIGHTBOTTOM].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag c b)")==0)
    {
      m_cFlags [FLAG_CENTERBOTTOM].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  
  //  Erkennen der Strafraumflaggen;
  if (strcasecmp (marker, "(flag p l b)")==0)
    {
      m_cFlags [FLAG_PENALTY_LEFTBOTTOM].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag p l c)")==0)
    {
      m_cFlags [FLAG_PENALTY_LEFTCENTER].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag p l t)")==0)
    {
      m_cFlags [FLAG_PENALTY_LEFTTOP].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag p r t)")==0)
    {
      m_cFlags [FLAG_PENALTY_RIGHTTOP].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag p r c)")==0)
    {
      m_cFlags [FLAG_PENALTY_RIGHTCENTER].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(flag p r b)")==0)
    {
      m_cFlags [FLAG_PENALTY_RIGHTBOTTOM].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  //  Erkennen der unbekannten Flagge;
  if (strcasecmp (marker, "(flag)")==0)
    {
      m_cFlags [FLAG_ANONYMOUS].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  
  //  Erkennen der Linien;
  if (strcasecmp (marker, "(line l)")==0)
    {
      m_cLines [LINE_LEFT].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(line t)")==0)
    {
      m_cLines [LINE_TOP].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(line r)")==0)
    {
      m_cLines [LINE_RIGHT].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(line b)")==0)
    {
      m_cLines [LINE_BOTTOM].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(line)")==0)
    {
      m_cLines [LINE_ANONYMOUS].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  if (strcasecmp (marker, "(line c)")==0)
    {
      m_cLines [LINE_CENTER].SetAll (dist, angle, distchange, dirchange);
      return;
    }
  // Rest wird ignoriert;
  return;
}

void Sensors::SeeBall (float dist, float angle, float distchange, float dirchange)
{
  m_cBall.SetAll (dist, angle, distchange, dirchange);
}

void Sensors::SeePlayer (float dist, float angle, float distchange, float dirchange, char* team, int number)
{
  Player* pl;
  pl = new Player (team, number);
  pl->SetAll (dist, angle, distchange, dirchange);
  m_cPlayers.Last ();
  m_cPlayers.InsertRight (pl);
}






