#include<stdio.h>
#include <stream.h>
#include"spielfe.h"

Spielfeld spielfeld; // unser Spielfeld

const Koerper Spielfeld::feldbreite = 68; // Ausdehnung in y-Richtung
const Koerper Spielfeld::feldlaenge = 105; // Ausdehnung in x-Richtung
const Koerper Spielfeld::strafraumbreite = 40.32; // Ausdehnung in y-Richtung
const Koerper Spielfeld::strafraumlaenge = 16.5; // Ausdehnung in x-Richtung

const Flagge Spielfeld::flaggen[anzahl_flaggen] = {
	{flagge_links_oben, {-feldlaenge / 2, feldbreite / 2}},
	{flagge_mitte_oben, {0, feldbreite / 2}},
	{flagge_rechts_oben, {feldlaenge / 2, feldbreite / 2}},
	{flagge_links_unten, {-feldlaenge / 2, -feldbreite / 2}},
	{flagge_mitte_unten, {0, -feldbreite / 2}},
	{flagge_rechts_unten, {feldlaenge / 2, -feldbreite / 2}},
	{flagge_s_links_oben, {-feldlaenge / 2 + strafraumlaenge, strafraumbreite / 2}},
	{flagge_s_links_mitte, {-feldlaenge / 2 + strafraumlaenge, 0}},
	{flagge_s_links_unten, {-feldlaenge / 2 + strafraumlaenge, -strafraumbreite / 2}},
	{flagge_s_rechts_oben, {feldlaenge / 2 - strafraumlaenge, strafraumbreite / 2}},
	{flagge_s_rechts_mitte, {feldlaenge / 2 - strafraumlaenge, 0}},
	{flagge_s_rechts_unten, {feldlaenge / 2 - strafraumlaenge, -strafraumbreite / 2}}
	};

const Feldlinie Spielfeld::linien[anzahl_linien] = {
	{obere_feldlinie, {-feldlaenge / 2,feldbreite / 2}, {feldlaenge / 2,feldbreite / 2}},
	{untere_feldlinie,  {-feldlaenge / 2,-feldbreite / 2}, {feldlaenge / 2,-feldbreite / 2}},
	{linke_feldlinie, {-feldlaenge / 2,feldbreite / 2}, {-feldlaenge / 2,-feldbreite / 2}},
	{rechte_feldlinie, {feldlaenge / 2,feldbreite / 2}, {feldlaenge / 2,-feldbreite / 2}},
	{mittlere_feldlinie, {0,feldbreite / 2}, {0,-feldbreite / 2}}
	};

const Tor Spielfeld::tore[anzahl_tore] = 
	{{rechtes_tor, {feldlaenge / 2,0}},{linkes_tor, {-feldlaenge / 2,0}}};

void Koordinate::ausgeben(void)
	{
	printf("(x,y) = ( %lf , %lf ), Richtung ist %lf Grad.\n",p.x1,p.x2,rad_in_deg(richtung));
	}

void Relative_Koordinate::ausgeben(void)
	{
	printf("Abstand ist %lf , rel. Richtung ist %lf Grad.\n",d,rad_in_deg(richtung));
	}

double Spielfeld::computeFaceDir( const Feldlinie &linie1,const Feldlinie &linie2,
	const Relative_Koordinate &delta_linie1,const Relative_Koordinate &delta_linie2,int count ) const  
{
  double angle = 0;
  int lineId = 0;
  // zuerst die Linie auswaehlen, deren Winkel ausgewertet wird.
  
  switch ( count ) {
    case 0:
      exception("Cannot compute face dir. Number of seen lines is 0.");
      break;
    case 1:
      // Position ist innerhalb des Spielfeldes.
      angle = delta_linie1.richtung;
      lineId = linie1.nummer;
      break;
    default:
      // Position ausserhalb des Spielfelds
      // Die zweite Linie wird von innen gesehen, daher
      // kann der Winkel berechnet werden als waere man im inneren
      // des Feldes.
      int i;
      if( delta_linie1.d > delta_linie2.d ) {
      	angle = delta_linie1.richtung;	
      	lineId = linie1.nummer;
      }
      else {
      	angle = delta_linie2.richtung;	
      	lineId = linie2.nummer;	
      }
      break;
  }
  // Jetzt abhaengig von der LinienId die absolute Blickrichtung berechnen 
  // cout << "(Gewaehlte Linie: " << lineId << ", " << angle;    
  switch ( lineId ) {
    case obere_feldlinie:
      if ( angle >= 0 ) {
	return - angle;
      }
      else {
	return -PI -angle;
      } 
      break;
    case untere_feldlinie:
      if ( angle > 0 ) {
	return  PI - angle;
      }
      else {
	return - angle;
      }
      break;
    case linke_feldlinie:
      if ( angle > 0 ) {
	return -PI/2 - angle;
      }
      else if ( angle < 0 ) {
	return PI/2 - angle;
      } 
      else {
	exception( "Couldn't compute face dir. Special case 1 is not treated." );
      }
      break;
    case rechte_feldlinie:
      if ( angle > 0 ) {
	return PI/2 - angle;
      }
      else if ( angle < 0 ) {
	return -PI/2 - angle;
      } 
      else {
      	exception( "Couldn't compute face dir. Special case 2 is not treated." );
      }
      break;
    default:
      exception( "Unknown line id in computeFaceDir" );
      break;
  }
}

int Spielfeld::triangulateViewPoint( const Feldlinie &linie1,const Feldlinie &linie2,
	const Relative_Koordinate &delta_linie1,const Relative_Koordinate &delta_linie2,
	int linien_count,
	const int id[10],const Relative_Koordinate delta_p[10],
	int punkt_count, Koordinate &S) const
{
  double angle;
  if (linien_count==0) return 0;
  angle = computeFaceDir(linie1,linie2,delta_linie1,delta_linie2,linien_count);
  
  // Schwerpunkt berechnen von 4 Positionen, die jeweils aus einer
  // Flagge und dem Blickwinkel bestimmt werden.
  // Gewichtet wird mit dem Reziproken der Entfernung der Flagge.
  int flagCount = punkt_count;
  Vektor pos;
  Vektor average;
  Vektor weightedPositionsSum;
  weightedPositionsSum.x1=0;
  weightedPositionsSum.x2=0;
  double sumOfWeights = 0;
  double weight;
 
  int i;
  for( i = 0; i < flagCount; i++ ) {
    pos = computePosFromFlag( id[i], delta_p[i], angle );

    weight = 1 / delta_p[i].d;
    
    weightedPositionsSum += (weight * pos);
    sumOfWeights += weight;
  }
  average = (1/sumOfWeights) * weightedPositionsSum;

  S.p=average;
  S.richtung=angle;
  return 1;
}

Vektor Spielfeld::computePosFromFlag( const int &id,const Relative_Koordinate &delta_p,
				  double faceDir ) const
{
  Vektor relPos;
  Vektor absPos;
  Vektor pPos;
  
  relPos.setzen( -(faceDir + delta_p.richtung), delta_p.d );
  
  if (id<anzahl_flaggen){
    absPos = Spielfeld::flaggen[id].p - relPos;
  }
  else if ( id >= 100 && id-100 < anzahl_tore ) {
    absPos = Spielfeld::tore[id-100].p - relPos;
  }
  return absPos;
}


int Spielfeld::berechne_relative_koordinate(const Feldlinie &linie,const Koordinate &S,Relative_Koordinate &delta) const
	// die exakten Vergleiche der Flieskommazahlen mussen durch Vorgeben einer Genauigkeit durch Ungefahre ersetzt werden.
	{
	Koord delta_y,delta_x,x,y;
	Winkel richtung;
	int ret;
	
	if(linie.nummer == untere_feldlinie || linie.nummer == obere_feldlinie)
		{// Trivialfalle ausschliesen (0,90,180,270 Grad)!
		if(ist_ungefaehr(S.richtung,0) || ist_ungefaehr(S.richtung,-PI) || 
			ist_ungefaehr(S.richtung,PI)) return 0; // Linie wird nicht gesehen.
		else if(ist_ungefaehr(S.richtung,PI / 2))
			{
			if(linie.nummer == obere_feldlinie) return 0; // Linie nicht gesehen
			delta.richtung = PI / 2;
			delta.d = betrag(S.p.x2 - linie.p1.x2);
			return 1;	// Sieht orthogonal auf die Linie
			}
		else if(ist_ungefaehr(S.richtung,-PI / 2))
			{
			if(linie.nummer == untere_feldlinie) return 0; // Linie nicht gesehen
			delta.richtung = PI / 2;
			delta.d =  betrag(S.p.x2 - linie.p1.x2);
			return 1;	// Sieht orthogonal auf die Linie
			}
		
		delta_y = linie.p1.x2 - S.p.x2;
		delta_x = delta_y / tan(-S.richtung);
		x = S.p.x1 + delta_x; // x-Koordinate des Schnittpunktes der Sichtlinie mit der Feldlinie
		y = S.p.x2;
		delta.richtung = atan(delta_y / delta_x);
		if(x < linie.p1.x1 || x > linie.p2.x1) ret = 0;
		}
	if(linie.nummer == linke_feldlinie || linie.nummer == rechte_feldlinie)
		{
		if(ist_ungefaehr(S.richtung,-PI / 2) || ist_ungefaehr(S.richtung,PI / 2)) return 0; // Linie wird nicht gesehen.
		else if(ist_ungefaehr(S.richtung,0))
			{
			if(linie.nummer == linke_feldlinie) return 0; // Linie nicht gesehen
			delta.richtung = PI / 2;
			delta.d = betrag(S.p.x1 - linie.p1.x1);
			return 1;	// Sieht orthogonal auf die Linie
			}
		else if(ist_ungefaehr(S.richtung,PI) || ist_ungefaehr(S.richtung,-PI))
			{
			if(linie.nummer == rechte_feldlinie) return 0; // Linie nicht gesehen
			delta.richtung = PI / 2;
			delta.d =  betrag(S.p.x1 - linie.p1.x1);
			return 1;	// Sieht orthogonal auf die Linie
			}
		
		delta_x = linie.p1.x1 - S.p.x1;
		delta_y = tan(-S.richtung) * delta_x;
		y = S.p.x2 + delta_y; // y-Koordinate des Schnittpunktes der Sichtlinie mit der Feldlinie
		delta.richtung = -atan(delta_x / delta_y);
		if(y < linie.p1.x2 || y > linie.p2.x2) ret = 0;
		}	

	delta.d = norm(delta_x,delta_y);
	return ret;
	}

int Spielfeld::berechne_relative_koordinate(const Punkt &p,const Koordinate &S,Relative_Koordinate &delta) const
	{
	Koord delta_y = p.x2 - S.p.x2,delta_x = p.x1 - S.p.x1;
	
	delta.d = norm(delta_x,delta_y);
	if (ist_ungefaehr (p.x2, S.p.x2))
	{
      if (delta_x >=0)
        delta.richtung = -S.richtung;
      else
        if (S.richtung < 0)
          delta.richtung = -PI-S.richtung;
        else
          delta.richtung = PI-S.richtung;
   }
	else
   {
// Ubrige Falle
		if(delta_x > 0 && delta_y > 0)
			delta.richtung = -S.richtung - atan(delta_y / delta_x);
		else if(delta_x > 0 && delta_y < 0)
			delta.richtung = -atan(delta_y / delta_x) - S.richtung;
		else if(delta_x < 0 && delta_y < 0)
			delta.richtung = PI - atan(delta_y / delta_x) - S.richtung;
		else if(delta_x < 0 && delta_y > 0)
			delta.richtung = -PI - atan(delta_y / delta_x) - S.richtung;
		
		if(delta.richtung > PI) delta.richtung -= 2 * PI;
		else if(delta.richtung < -PI) delta.richtung += 2 * PI;
		}

	return 0;
	}

int Spielfeld::transformiere(RelPos &rp,Relative_Koordinate &rk)
	{
	if(rp.GetDistance() == DIST_NOTVALID || rp.GetAngle() == ANGLE_NOTVALID) return 0;
	rk.d = rp.GetDistance();
	rk.richtung = deg_in_rad(rp.GetAngle());
	return 1;
	}

void Spielfeld::transformiere(Relative_Koordinate &rk,RelPos &rp)
	{
	rp.SetDistance(rk.d);
	rp.SetAngle(rad_in_deg(rk.richtung));
	}
